import json
import shutil
import subprocess
import tempfile
from abc import ABC, abstractmethod
import os
import copy
from contextlib import contextmanager
from functools import lru_cache, wraps
from pathlib import Path
from typing import Callable, Tuple, List, Union
from warnings import warn
import pathos.multiprocessing as mp
import time
import platform

import pandas as pd
from pandas import DataFrame as DF
import numpy as np
import platypus

import config
from problem import Problem
from objectives import read_eso


def _list_cache(enabled=config.cache['enabled'], size=config.cache['size'],
                typed=config.cache['typed']):
    """Decorator that allows the lru_cache (last recently used cache)
    to work on functions that receive lists"""

    def _list_cache_decorator(f):
        if enabled:
            cache_f = lru_cache(size, typed)(f)

            @wraps(cache_f)
            def safe_f(individual, *args, **kwargs):
                return cache_f(tuple(individual), *args, **kwargs)

            safe_f.cache_clear = cache_f.cache_clear
            return safe_f
        # add a no-op as the cache clear operation so it does not error if disabled in the config
        f.cache_clear = lambda: None
        return f

    return _list_cache_decorator


@contextmanager
def working_directory(path=None):
    old_wd = os.getcwd()
    if path is None:
        tempdir = tempfile.TemporaryDirectory()
        path = tempdir.name
    else:
        tempdir = None
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(old_wd)
        if tempdir is not None:
            tempdir.cleanup()


def run(building, schema=config.files['data dict'], epw=config.files['epw'],
        out_dir=config.out_dir, err_dir=config.err_dir, error_mode='Silent'):
    def resolve(path):
        if path is None:
            return None
        return Path(path).resolve()

    schema, epw, out_dir, err_dir = (resolve(p) for p in (schema, epw, out_dir, err_dir))
    with working_directory(out_dir) as out_dir:
        try:
            building_path = 'in.idf'
            building.saveas(building_path)
        except AttributeError:
            building_path = str(Path(out_dir, 'in.epJSON'))
            with open(building_path, 'w') as f:
                json.dump(building, f)
        try:
            if platform.system() is 'Windows':
                subprocess.run(['energyplus', '--idd', str(schema),
                                '--weather', str(epw), building_path], check=True, shell=True)
            else:
                subprocess.run(['energyplus', '--idd', str(schema),
                                '--weather', str(epw), building_path], check=True, shell=False)
            results = read_eso(out_dir)
            return results
        except subprocess.CalledProcessError as e:
            if err_dir is not None and Path(out_dir).resolve() != Path(err_dir).resolve():
                if os.path.exists(err_dir):
                    shutil.rmtree(err_dir)
                shutil.copytree(out_dir, err_dir)
                if error_mode != 'Silent':
                    filename = str(Path(err_dir, 'eplusout.err'))
                    file = open(filename, "r")
                    for line in file:
                        print(line)
                    print()
            raise e


class AbstractEvaluator(ABC):
    """Base class for Evaluators.

    This template requires that Evaluators are callable.
    It also gives them the df_apply method and result caching."""

    error_mode_options = {'Failfast', 'Silent', 'Print'}

    def __init__(self, problem: Problem,
                 error_mode='Failfast', error_value: tuple = None, multi = False):
        """

        :param problem: description of the inputs and outputs the evaluator will use
        :param error_mode: One of {'Failfast', 'Silent', 'Print'}.
            Failfast: Any error aborts the evaluation.
            Silent: Evaluation will return the `error_value` for any lists of values that raise an error.
            Print: Same as silent, but warnings are printed to stderr for any errors.
        :param error_value: The value of the evaluation if an error occurs. Incompatible with error_mode='Failfast'.
            must have the form (objective_values, constraint_values).
        :param multi: whether or not to use multiprocessing, using this disables cache
        """
        self.problem = problem
        self.error_mode = error_mode
        self.error_value = error_value
        self.validate_error_mode()
        self.multi = multi
        # this makes the cache instance-specific instead of class-specific
        if not self.multi:
            self._call = _list_cache()(self._uncached_call)

    # TODO: Accept more flexible input formats
    # Factoring out some of the code from _call related to handling different tuple formats might be of use
    def validate_error_mode(self):
        msg = f'Invalid error mode, only {self.error_mode_options} are allowed'
        assert self.error_mode in self.error_mode_options, msg
        if self.error_mode == 'Failfast':
            assert self.error_value is None, 'error value cannot be set when in Failfast mode'
            return
        # intuit error_value if needed
        if self.error_value is None:
            self.error_value = (None, None)
        err_out, err_constraint = self.error_value
        if err_out is None:
            err_out = tuple(float('inf') if minimize else float('-inf')
                            for minimize in self.problem.minimize_outputs)
        if err_constraint is None:
            # TODO: Intuit constraint error values as well
            err_constraint = ()
        msg = 'error value must match outputs length'
        assert len(err_out) == self.problem.num_outputs, msg
        msg = 'error value must match constraints length'
        assert len(err_constraint) == self.problem.num_constraints, msg
        self.error_value = (err_out, err_constraint)
        self.error_mode = self.error_mode

    @abstractmethod
    def eval_single(self, values: list, **kwargs) -> Tuple[Tuple, Tuple]:
        """Returns the objective results for a single list of parameter values.

        :param values: A list of values to set each parameter to,
            in the same order as this evaluator's inputs
        :param kwargs: Any keyword arguments
        :return: a tuple of the objectives and constraints
        """
        pass

    def _uncached_call(self, values: list, separate_constraints=None, **kwargs) -> tuple:
        # This method is a workaround to allow lru_cache to be instance-specific and work on __call__
        if separate_constraints is None:
            separate_constraints = self.problem.num_constraints > 0
        try:
            self.validate(values)
            outputs, constraints = self.eval_single(values, **kwargs)
        except Exception as e:
            if self.error_mode != 'Silent':
                msg = ''
                if self.problem.inputs is not None:
                    msg += f'for inputs: {self.problem.names("inputs")} '
                msg += f'problematic values were: {values}'
                warn(msg)
            if self.error_mode == 'Failfast':
                raise e
            else:
                outputs, constraints = self.error_value
        if separate_constraints:
            return outputs, constraints
        return outputs + constraints

    def __call__(self, values: list, separate_constraints=None, **kwargs) -> tuple:
        """Returns the objective results for a single list of parameter values.

        :param values: A list of values to set each parameter to,
            in the same order as this evaluator's inputs
        :param separate_constraints: which output format to use:
            objectives and constraints are both tuples of the measured values
            True: (objectives, constraints)
            False: objectives + constraints
            None: Same as True if there is at least one constraint, else the same as False
        :return: a tuple of the objectives' results
        """
        # Enables validation and caching in subclasses
        # Redirects calls to evaluate a list of values to the eval_single function.
        # Override eval_single, not this method.
        # Values can be empty to allow evaluating at the current state, but this is not the default behaviour
        # If this is not supported, the validate function of a subclass should reject an empty list
        if self.multi:
            return self._uncached_call(values, separate_constraints, **kwargs)
        else:
            return self._call(values, separate_constraints, **kwargs)

    def cache_clear(self):
        """Clears any cached vales of calls to this evaluator.
        This should be called whenever the evaluator's outputs could have changed."""
        if not self.multi:
            self._call.cache_clear()

    def validate(self, values):
        """Takes a list of values and checks that they are a valid input for this evaluator."""
        if len(values) != self.problem.num_inputs:
            raise ValueError(f'Wrong number of input values.'
                             f'{len(values)} provided, {self.problem.num_inputs} expected')
        for p, value in zip(self.problem.inputs, values):
            if not p.validate(value):
                raise ValueError(f'Invalid value {value} for parameter {p}')

    def estimate_time(self, df: DF, processes: int = 4, **kwargs):
        """Prints out a very rough estimate of the amount of time a job will take to complete.
        Will underestimate smaller sample sets but becomes more accurate as they become larger.

        :param df: a DataFrame where each row represents valid input values for this Evaluator.
        :param processes: amount of cores to use (only relevent when using multiprocessing)
        """
        start = time.time()
        res = self(df.iloc[0])
        end = time.time()

        if self.multi and processes > 1:
            estimate = ((end - start) * df.shape[0] / processes)
        else:
            estimate = (end - start) * df.shape[0]
        h = np.floor(estimate / 3600)
        m = np.floor((estimate % 3600) / 60)
        s = np.floor((estimate % 3600) % 60)
        print(f"Estimated time for completion: {h} hours, {m} minutes, {s} seconds")

    def df_apply(self, df: DF, keep_input = False, processes: int = 4, **kwargs):
        """Applies this evaluator to an entire dataFrame, row by row.

        :param df: a DataFrame where each row represents valid input values for this Evaluator.
        :param keep_input: whether to include the input data in the returned DataFrame
        :param processes: amount of cores to use (only relevent when using multiprocessing)
        :return: Returns a DataFrame with one column containing the results for each objective.
        """

        def process(df):
            res = df.apply(self, axis=1, result_type='expand', separate_constraints=False, **kwargs)
            return res

        if self.multi == True and processes > 1:
            if(config.out_dir != None):
                raise ValueError("out_dir is set to something other than 'None', running in parallel will not work.")
            if df.shape[0] < processes:
                processes = df.shape[0]
            pool = mp.Pool(processes = processes)
            split = np.array_split(df, processes)
            pool_res = pool.map(process, split)
            pool.close()
            pool.join()
            results = []
            for result in pool_res:
                results.append(result)
            results = pd.concat(results)
            result = results.rename({i: name for i, name in enumerate(self.problem.names('outputs'))}, axis=1)
        else:
            result = df.apply(self, axis=1, result_type='expand', separate_constraints=False, **kwargs)
            result = result.rename({i: name for i, name in enumerate(self.problem.names('outputs'))}, axis=1)
        if keep_input:
            result = pd.concat([df, result], axis=1)
        return result

    def to_platypus(self) -> platypus.Problem:
        """Converts this evaluator (and the underlying problem) to a platypus compatible format

        :return: A platypus Problem that can optimise over this evaluator
        """
        problem = self.problem.to_platypus()
        problem.function = self
        return problem


class EvaluatorSR(AbstractEvaluator):
    """Surrogate Model Evaluator

    This evaluator is a wrapper around a surrogate model, as defined by a function.
    """
    eval_func_format = Callable[[List], Tuple[float, ...]]

    def __init__(self, evaluation_func: eval_func_format,
                 problem: Problem, error_mode='Failfast', error_value=None, multi=False):
        """
        :param evaluation_func: a function that takes as input an list of values,
                and gives as output a tuple of the objective values for that point in the solution space
        :param problem: description of the inputs and outputs the evaluator will use
        :param multi: whether or not to use multiprocessing, using this disables cache
        """

        super().__init__(problem=problem, error_mode=error_mode, error_value=error_value, multi=multi)
        self._evaluation_func = evaluation_func

    def eval_single(self, values: List) -> Tuple:
        return self._evaluation_func(values)


tabular = Union[DF, np.array]


# TODO: Add an option/subclass that automatically bundles several single variable models into a multiobjective model.
class AdaptiveSR(AbstractEvaluator, ABC):
    """A Template for making adaptive sampling based models compatible with the evaluator interface."""

    # helper functions provided by AdaptiveSR (Generally avoid editing these, but use them as needed)
    # append_data(X, y)
    # do_infill
    # get_from_reference

    # functions with defaults (These can be removed from this template if you like the defaults)
    # They may depend on some of the optional functions in order to work if using the defaults
    # __init__
    # infill -> get_infill, _update_model
    # update_model

    # optional functions (These will not work unless you implement them)
    # get_infill

    # required functions
    # train
    # eval_single

    def __init__(self, reference: AbstractEvaluator = None,
                 error_mode='Failfast', error_value=None, multi=False):
        self.reference: AbstractEvaluator = reference
        super().__init__(problem=reference.problem, error_mode=error_mode, error_value=error_value, multi=multi)

        self.model = None
        self.data: DF = pd.DataFrame(columns=self.problem.names(parts=['inputs', 'outputs', 'constraints']))

    @property
    def problem(self):
        return self.reference.problem

    @problem.setter
    def problem(self, value: Problem):
        self.reference.problem = value

    def append_data(self, data: tabular, deduplicate=True) -> None:
        """Adds the X and y data to input_data and output_data respectively

        :param data: a table of training data to store
        :param deduplicate: whether to remove duplicates from the combined DataFrame
        :return:
        """
        self.cache_clear()  # TODO: decide on a consistent way of tracking this
        # can we assume users will only modify the data using this method or will call cache_clear themselves
        new_data = self.problem.to_df(data, ['inputs', 'outputs', 'constraints'])
        self.data = self.data.append(new_data, ignore_index=True)
        if deduplicate:
            self.data.drop_duplicates(inplace=True)

    def get_infill(self, num_datapoints: int) -> tabular:
        """Generates data that is most likely to improve the model, and can be used for retraining.

        :param num_datapoints: the number of datapoints to generate
        :return: the datapoints generated, in some tabular datastructure
        """
        raise NotImplementedError

    def do_infill(self, data: DF) -> None:
        """Updates the model using the inputs X and outputs y, and stores the added data

        :param data: a table of training data
        :return: None
        """
        old_df = self.data
        df, parts = self.problem.partial_df(data, parts=['inputs', 'outputs', 'constraints'])
        if parts == ['inputs']:
            outputs: DF = self.get_from_reference(df)
            df = pd.concat([df, outputs], axis=1)
        self.append_data(df)
        if self.model is None:
            self.train()
        else:
            self.update_model(df, old_df)

    def update_model(self, new_data: tabular, old_data: DF = None) -> None:
        """Modifies self.model to incorporate the new data.

        This function should not edit the existing data

        :param new_data: a table of inputs and outputs
        :param old_data: the table of inputs and outputs without the new data
        :return: None
        """
        self.train()

    def infill(self, num_datapoints: int) -> None:
        """Adds num_datapoints samples to the model and updates it.

        :param num_datapoints: number of datapoints to add to the model's training set
        :return: None
        """
        inputs: DF = self.problem.to_df(self.get_infill(num_datapoints), 'inputs')
        outputs: DF = self.get_from_reference(inputs)
        self.do_infill(pd.concat([inputs, outputs], axis=1))

    @abstractmethod
    def train(self) -> None:
        """Generates a new model using the stored data, and stores it as self.model"""
        pass

    @abstractmethod
    def eval_single(self, values: List, **kwargs) -> Tuple:
        """Evaluates a single input point

        :param values: The datapoint to evaluate
        :param kwargs: Arbitrary keyword arguments.
        :return: A tuple of the predicted outputs for this datapoint
        """
        pass

    def get_from_reference(self, X: tabular) -> DF:
        """Use the reference evaluator to get the real value of a dataframe of datapoints

        :param X: a table containing the datapoints to evaluate
        :return: a DataFrame containing the results of the datapoints
        """
        df = self.problem.to_df(X, 'inputs')
        return self.reference.df_apply(df)


class EvaluatorEP(AbstractEvaluator, ABC):
    """This evaluator uses a Problem to modify a building, and then simulate it.
    It keeps track of the building and the weather file."""

    def __init__(self, problem: Problem,
                 building, epw=config.files['epw'], out_dir=config.out_dir,
                 err_dir=config.err_dir, error_mode='Failfast', error_value=None, multi=False):
        """
        :param problem: a parametrization of the building and the desired outputs
        :param building: the building that is being simulated.
        :param epw: the epw file representing the weather
        :param out_dir: the directory used for files created by the EnergyPlus simulation.
        :param err_dir: the directory where files from a failed run are stored.
        :param error_mode: One of {'Failfast', 'Silent', 'Print'}.
            Failfast: Any error aborts the evaluation.
            Silent: Evaluation will return the `error_value` for any lists of values that raise an error.
            Print: Same as silent, but warnings are printed to stderr for any errors.
        :param error_value: The value of the evaluation if an error occurs. Incompatible with error_mode='Failfast'.
        :param multi: whether or not to use multiprocessing, using this disables cache

        """
        super().__init__(problem=problem, error_mode=error_mode, error_value=error_value, multi=multi)
        self.out_dir = out_dir
        self.err_dir = err_dir
        for io in self.problem:
            io.setup(building)
        self.building = building
        self.epw = epw

    def df_apply(self, df: DF, keep_input = False, processes: int = 4, **kwargs):
        """Applies this evaluator to an entire dataFrame, row by row.

        :param df: a DataFrame where each row represents valid input values for this Evaluator.
        :param keep_input: whether to include the input data in the returned DataFrame
        :param processes: amount of cores to use (only relevent when using multiprocessing)
        :return: Returns a DataFrame with one column containing the results for each objective.
        """

        # Helper function for multiprocessing
        def process(df):
            res = df.apply(self, axis=1, result_type='expand', separate_constraints=False, **kwargs)
            return res

        if self.multi == True and processes > 1:
            if(config.out_dir != None):
                raise ValueError("out_dir is set to something other than 'None', running in parallel will not work.")
            if df.shape[0] < processes:
                processes = df.shape[0]
            pool = mp.Pool(processes = processes)
            split = np.array_split(df, processes)
            pool_res = pool.map(process, split)
            pool.close()
            pool.join()
            results = []
            for result in pool_res:
                results.append(result)
            results = pd.concat(results)
            result = results.rename({i: name for i, name in enumerate(self.problem.names('outputs'))}, axis=1)
        else:
            result = df.apply(self, axis=1, result_type='expand', separate_constraints=False, **kwargs)
            result = result.rename({i: name for i, name in enumerate(self.problem.names('outputs'))}, axis=1)

        # Fixes Dataframe output when using Multiply or Add mode
        j = 0
        for parameter in self.problem.inputs:
            if not isinstance(parameter, int):
                if parameter.mode is not 'Set':
                    names = self.problem.names('inputs')
                    objects = parameter.selector.get_objects(self.building)
                    values = parameter.selector.get(self.building)
                    if parameter.mode is 'Multiply':
                        df.rename(columns={names[j] : names[j] + ' Multiplier'}, inplace=True)
                    if parameter.mode is 'Add':
                        df.rename(columns={names[j] : names[j] + ' Addend'}, inplace=True)
            j += 1
        if keep_input:
            result = pd.concat([df, result], axis=1)
        return result

    def eval_single(self, values: list):
        current_building = copy.deepcopy(self.building)
        for p, value in zip(self.problem.inputs, values):
            # apply the modification of each parameter to the building using the corresponding value
            p.transformation_function(current_building, value)
        results = run(current_building, out_dir=self.out_dir, err_dir=self.err_dir, epw=self.epw, error_mode=self.error_mode)
        outputs = tuple((objective(results) for objective in self.problem.outputs))
        constraints = tuple((constraint(results) for constraint in self.problem.constraints))
        return outputs, constraints

    @property
    def building(self):
        return self._building

    @building.setter
    def building(self, value):
        """Changes the building simulated.
        Changing this resets the cache.

        :param value: the building to use
        :return: None
        """
        self.cache_clear()
        self._building = value

    @property
    def epw(self):
        return self._epw

    @epw.setter
    def epw(self, value: str) -> None:
        """Changes the epw file used with this building.
        Changing this resets the cache.

        :param value: path to the new epw file to use.
        :return: None
        """
        self.cache_clear()
        self._epw = value



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

from besos.optimizer import rbf_opt
from besos.evaluator import EvaluatorSR
from besos.parameters import Parameter, RangeParameter
from besos.problem import Problem

# +
def obj_funct(x):
    return (x[0]*x[1] - x[2],),()

param_list = [Parameter(value_descriptor=RangeParameter(0,10)) for _ in range(3)]

problem = Problem(param_list, 1)

evaluator = EvaluatorSR(obj_funct, problem)
# -

rbf_opt(evaluator, 30)





# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # Optimisation Run Flexibility
# It is possible to change the configuration of the aglorithm part way through the optimisation process, or even to switch algorithms completely. Doing so requires using platypus algorithms directly, instead of the algorithm wrappers provided through the optimizations module.
#
# First we create an example problem.

# +
from besos import eppy_funcs as ef

from besos.parameters import expand_plist
from besos.evaluator import EvaluatorEP
from besos.problem import EPProblem
from besos import optimizer
import platypus

# +
idf = ef.get_idf()

parameters=expand_plist(
    {'NonRes Fixed Assembly Window':
     {'UFactor':(0.1,5),
      'Solar Heat Gain Coefficient':(0.01,0.99)
     },
     'Mass NonRes Wall Insulation':{'Thickness':(0.01,0.09)},
    })

objectives = ['Electricity:Facility', 'Gas:Facility']

problem = EPProblem(parameters, objectives)

evaluator = EvaluatorEP(problem, idf)
# -

# Here we set up the first algorithm. The `to_platypus` shortcut converts the evaluator to a `platypus.Problem` object.

platypus_problem = evaluator.to_platypus()
algorithm = platypus.NSGAII(problem=platypus_problem)

# Use the **stop button** at the top of the notebook to interrupt the following cell, since it will take a while.  
# Note: The output from the next cells will vary from run to run, due to the randomness used in platypus' algorithms, as well as the amount of time this cell runs for.

try:
    algorithm.run(1000)
except KeyboardInterrupt:
    print('Algorithm interrupted')
algorithm.population[:5]

# Now we want to continue from where the first algorithm left off, but with `EpsMOEA` and only 10 evaluations. In order to make the population carry over, we use the `InjectedPopulation` generator. Then we run the second algorithm.
#
# If we had let the first algorithm finish, we could use `algorithm.result` instead of `algorithm.population` to use the solutions found by the first algorithm as a starting point for the next.

generator = platypus.InjectedPopulation(algorithm.population)
alg2 = platypus.EpsMOEA(problem=platypus_problem, generator=generator, epsilons=3, population_size=10)
alg2.run(10)

# Now we convert the solutions that were found to a dataframe and display them.

optimizer.solutions_to_df(alg2.result, problem, parts=['inputs', 'outputs'])



import pytest

import eppy_funcs as ef
from evaluator import EvaluatorEP, EvaluatorSR
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector, expand_plist
from problem import EPProblem, Problem
import sampling


@pytest.fixture
def building():
    return ef.get_building()
    #return ef.get_building('D:/EnergyPlusV9-0-1/ExampleFiles/SingleFamilyHouse_TwoSpeed_CutoutTemperature.idf')


@pytest.fixture
def parameters():
    parameters = [
        #Parameter(FieldSelector(object_name='HOUSE OCCUPANCY',
        #        field_name='Field 4'),
        #        value_descriptor=CategoryParameter([1.0]))
        Parameter(FieldSelector(object_name='NonRes Fixed Assembly Window',
                field_name='Solar Heat Gain Coefficient'),
                value_descriptor=RangeParameter(0.01,0.99))
        ]
    return parameters


@pytest.fixture
def problem(parameters):
    objectives = ['Electricity:Facility']
    problem = EPProblem(parameters, objectives)
    return problem


def test_eval(building, problem):
    print(type(building))
    evaluator = EvaluatorEP(problem, building)
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, 20)

    outputs = evaluator.df_apply(samples, keep_input=True)
    print()
    print(outputs)
    assert 0



import contextlib
import numpy as np

import config
from eppy.modeleditor import IDF
import json

from functools import partial
import warnings

from errors import ModeError

"""This file is a collection of functions to interact with eppy.
They could do with some sorting and refactoring."""


def convert_format(s: str, place, mode):
    """Converts energyPlus names to their Eppy equivalents

    :param s: name to convert
    :param place: the type of place that is being used 'field' or 'class'
    :param mode: whether to convert to idf or json formating
    :return: the converted name
    """
    # TODO: Find an authoritative source for the naming conventions
    if mode == 'idf':
        if place == 'field':
            return s.replace(' ', '_').replace('-', '')
        if place == 'class':
            return s.upper()
    if mode == 'json':
        if place == 'field':
            return s.replace(' ', '_').replace('-', '_').lower()
        if place == 'class':
            # uses camel case, s.title() removes some of the capitals and does not work
            return s
    raise ModeError(message=f'no format defined for place:{place} and mode:{mode}')


def get_mode(building):
    if isinstance(building, IDF):
        return 'idf'
    if isinstance(building, dict):
        return 'json'
    raise ModeError(message=f'Cannot find a valid mode for {building}')


def get_idf(idf_file: str = config.files.get('idf'),
            idd_file: str = config.files.get('idd'),
            output_directory=config.out_dir) -> IDF:
    """Uses eppy to read an idf file and generate the corresponding idf object"""
    # Trying to change the idd file inside a program will cause an error
    IDF.setiddname(idd_file)
    # TODO: Fix this rather than hiding it.
    # calling IDF causes a warning to appear, currently redirect_stdout hides this.
    with contextlib.redirect_stdout(None):
        idf = IDF(idf_file)
        # override the output location so I stop messing up
        idf.run = partial(idf.run, output_directory=output_directory)
        return idf


def get_building(building=None,
                 data_dict=None,
                 output_directory=config.out_dir,
                 mode=None):
    if mode is None and building is None:
        building = config.files.get('building')
        mode = config.energy_plus_mode
    if mode is None:
        ending = building.split('.')[-1]
        if ending == 'idf':
            mode = 'idf'
        elif ending == 'epJSON':
            mode = 'json'
        else:
            warnings.warn(f'expected "idf" or "epJSON" file. building has extension {ending}')
    if building is None:
        building = config.files[mode]
    if mode == 'idf':
        data_dict = data_dict or config.files['idd']
        return get_idf(idf_file=building, idd_file=data_dict, output_directory=output_directory)
    elif mode == 'json':
        with open(building) as f:
            return json.load(f)
    ModeError(mode)


def get_windows(building):
    mode = get_mode(building)
    if mode == 'idf':
        return ((window.Name, window) for window in building.idfobjects['FENESTRATIONSURFACE:DETAILED'])
    elif mode == 'json':
        return building['FenestrationSurface:Detailed'].items()
    else:
        raise ModeError(mode)


def wwr_all(building, wwr: float) -> None:
    """Sets the wwr for all walls that have a window.
    Will malfunction if there are multiple windows on one wall
    """
    mode = get_mode(building)
    windows = get_windows(building)
    for window_name, window in windows:
        wwr_single(window, wwr, building, mode)


def set_vertex(idfObj, vertexNum: int, x: float = 0, y: float = 0, z: float = 0):
    """Sets a single vertex of the passed idfObject (idfObj)
     to the specified x,y and z coordinates."""
    for val, name in zip((x, y, z), 'XYZ'):
        idfObj['Vertex_{}_{}coordinate'.format(vertexNum, name)] = round(val, 2)


def one_window(building):
    """Removes some windows so that each wall has no more than one"""
    mode = get_mode(building)
    walls = set()
    toRemove = []
    windows = get_windows(building)
    for window_name, window in windows:
        if mode == 'idf':
            wall_name = window.Building_Surface_Name
        elif mode == 'json':
            wall_name = window['building_surface_name']
        else:
            raise ModeError(mode)
        if wall_name in walls:
            toRemove.append((window_name, window))
        else:
            walls.add(wall_name)
    if mode == 'idf':
        for window_name, window in toRemove:
            building.idfobjects['FENESTRATIONSURFACE:DETAILED'].remove(window)
    elif mode == 'json':
        for window_name, window in toRemove:
            building['FenestrationSurface:Detailed'].pop(window_name)


def wwr_single(window, wwr: float, building, mode):
    """Sets `window` to have the window to wall ratio specified by `wwr`"""

    # will not work for some orientations
    # multiple windows on a single wall will break this
    if mode == 'idf':
        def coordinates(ax):
            return [window[f'Vertex_{n}_{ax.upper()}coordinate'] for n in range(1, 5)]

        wall = building.getobject('BUILDINGSURFACE:DETAILED', window.Building_Surface_Name)
    elif mode == 'json':
        def coordinates(ax):
            return [window[f'vertex_{n}_{ax.lower()}_coordinate'] for n in range(1, 5)]

        wall = building['BuildingSurface:Detailed'][window['building_surface_name']]
    else:
        raise ModeError(mode)

    xs = coordinates('X')
    ys = coordinates('Y')
    zs = coordinates('Z')

    if max(ys) == min(ys):
        axis = 'x'
        axs = xs
    elif max(xs) == min(xs):
        axis = 'y'
        axs = ys
    else:
        raise ValueError('The window is not aligned with the x or y axes')
    width = max(axs) - min(axs)
    scale_factor = np.sqrt(wwr)
    new_width = width * scale_factor
    height = max(zs) - min(zs)
    new_height = height * scale_factor

    startW = (width - new_width) / 2
    endW = startW + new_width
    startH = (height - new_height) / 2
    endH = startH + new_height

    # Maintains vertex order by mimicking the current order
    s = [0] * 4
    for vertex in range(0, 4):
        if zs[vertex] == max(zs):
            # vertex on the top
            if axs[vertex] == max(axs):
                # TOP RIGHT
                s[0] += 1
                set_vertex(window, vertex + 1, z=endH, **{axis: endW})
            else:
                # TOP LEFT
                s[1] += 1
                set_vertex(window, vertex + 1, z=endH, **{axis: startW})
        else:
            if axs[vertex] == max(axs):
                # BOTTOM RIGHT
                s[2] += 1
                set_vertex(window, vertex + 1, z=startH, **{axis: endW})
            else:
                # BOTTOM LEFT
                s[3] += 1
                set_vertex(window, vertex + 1, z=startH, **{axis: startW})
    assert s == [1] * 4, ('verticesS are wrong:', s)



import pyDOE2
import numpy as np
import pandas as pd

from typing import Callable, Any

from IO_Objects import Descriptor
from problem import Problem


_sampler = Callable[[int, int, Any], np.ndarray]  # the format of all sampler functions


def dist_sampler(sampler: _sampler, problem: Problem, num_samples: int, *args, **kwargs) -> pd.DataFrame:
    """uses the sampling function provided to generate `num_samples` sets of values
    These values will be valid for the corresponding inputs

    :param problem: Problem that the inputs should apply to
    :param sampler: a function that is used to produce the distribution of samples
    :param num_samples: number of samples to take

    :param args: Arguments passed to the sampling function
    :param kwargs: Arguments passed to the sampling function
    :return: pandas DataFrame with one column per parameter, and `num_samples` rows

    """
    samples = sampler(num_samples, problem.num_inputs, *args, **kwargs)
    data = {name: _sample_param(param, samples[:, i]) for i, (name, param)
            in enumerate(zip(problem.names('inputs'), problem.inputs))}
    df = pd.DataFrame(data)
    # enforce the correct order in case it was lost by the dictionary
    df = df[problem.names('inputs')]
    return df


def add_extremes(df, problem):
    """adds two datapoints with the minimum and maximum values for each attribute

    :param data:
    :param problem:
    :return:
    """
    new = dist_sampler(extremes, problem, 2)
    return pd.concat([df, new], ignore_index=True)


def extremes(samples: int, attributes: int=None) -> np.ndarray:
    if attributes is None:
        samples, attributes = 2, samples
    assert samples == 2, "extremes can only produce two samples"
    return np.array([np.zeros(attributes), np.ones(attributes)])


def _sample_param(parameter: Descriptor, values):
    try:
        # this will only work if p.sample is numpy compatible
        return parameter.sample(values)
    except TypeError:
        # apply non-vectorised version of p.sample
        return [parameter.sample(x) for x in values]


# all of the following return a 2d array of shape (samples, attributes)
# which contains values between 0 and 1


random: _sampler = np.random.rand


def seeded_sampler(samples: int, attributes: int, seed=0) -> np.ndarray:
    np.random.seed(seed)
    return np.random.rand(samples, attributes)


def lhs(samples: int, attributes: int, *args, **kwargs) -> np.ndarray:
    return pyDOE2.lhs(attributes, samples, *args, **kwargs)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# This notebook covers how to use the two simplest evaluators, the EnergyPlus Evaluator (`EvaluatorEP`) and the surrogate Evaluator (`EvaluatorSR`)

from besos.evaluator import EvaluatorEP, EvaluatorSR
from besos import eppy_funcs as ef
from besos.parameters import FieldSelector, Parameter
from besos.problem import EPProblem, Problem

# ## EvaluatorEP
# This evaluator needs a building, and a problem with parameters that can modify it, and objectives that correspond to measurements of the building.  
# Parameters that have a selector of the type `FieldSelector` or `FilterSelecor` will work, and `GenericSelector` could work, if the right functions are provided.  
# Objectives of the type `MeterReader` or `VariableReader` will work.

# +
building = ef.get_building()

parameters = [Parameter(FieldSelector(object_name='Mass NonRes Wall Insulation',
                                      field_name='Thickness'))]

objectives = ['Electricity:Facility', 'Gas:Facility'] # the default is just 'Electricity:Facility'
problem=EPProblem(parameters, objectives) #EPP Problem automatically converts these to MeterReaders

evaluator = EvaluatorEP(problem, building)

evaluator([0.5]) # run with thickness set to 0.5 
# -

# ## EvaluatorSR
# This evaluator needs a function that takes the correct number of inputs and produces outputs in the format `(objectives, constraints)` where objectives and constraints are tuples.  
# Since this evaluator only uses it's parameters to track the number of inputs and outputs (and their names), we can use numbered placeholders that are automatically generated by `Problem`

# +
def function(values):
    return ((values[0], values[0]**2), ())
    
print('function value at 4 is:',function([4]))

new_problem = Problem(1,2,0) # this denotes a problem which takes 1 input, produces 2 outputs and no
#contraints. The placeholder parameters/objectives will be generated automatically.

# These names are used on the headings of DataFrames generated by the problem and evaluators that use it
print(problem.names(), new_problem.names())

evaluator_1 = EvaluatorSR(function, problem)
evaluator_2 = EvaluatorSR(function, new_problem)
print(evaluator_1([4], separate_constraints=True))
print(evaluator_2([4], separate_constraints=True))
# -

# separate constraints=True uses the `(objectives, constraints)` format for output.
# It can also be set to `False` or `None`  
# `False`: concatenates the tuples  
# `None`: Same as True if there is at least one constraint, else the same as False  

evaluator_1([4])





# this error was created to avoid having to repeat the same code in every invalid mode check
class ModeError(ValueError):
    def __init__(self, mode=None, message=None):
        if message is None:
            message = f'Invalid mode {mode}. Expected "idf" or "json"'
        super().__init__(message)



import warnings
from typing import List, Dict, Tuple
from abc import ABC, abstractmethod

import platypus

import eppySupport
from IO_Objects import Descriptor, IOBase, Selector, DummySelector, AnyValue
from config import range_parameter as conf
import eppy_funcs as ef  # incorrectly detected as an unused import by pycharm
from errors import ModeError


# Descriptors

class RangeParameter(Descriptor):
    """Represents an aspect of a building that can take on value from an interval."""
    pandas_type = float
    rbf_type = 'R'

    def __init__(self, min_val=conf.get('min'),
                 max_val=conf.get('max')):
        super().__init__()
        if min_val > max_val:
            raise ValueError('minimum is larger than maximum')
        self.min = min_val
        self.max = max_val
        self._add_reprs(['min', 'max'])
        self.platypus_type = platypus.Real(self.min, self.max)

    def validate(self, val):
        min_ = float('-inf') if self.min is None else self.min
        max_ = float('inf') if self.max is None else self.max
        return min_ <= val <= max_

    def sample(self, value):
        """Transforms a value in [0, 1] into a value in [min, max].
        This transformation is uniform.

        :param value: a value in the interval [0, 1]
        :return: a value from the interval [min, max]
        """
        return (self.max - self.min) * value + self.min

    def __str__(self):
        return f'{self.__class__.__name__} [{self.min}, {self.max}]'


class Category(platypus.Subset):
    def __init__(self, elements):
        super().__init__(elements=elements, size=1)

    def encode(self, value):
        return [value]

    def decode(self, value):
        return value[0]


class CategoryParameter(Descriptor):
    pandas_type = 'category'

    def __init__(self, options: list, **kwargs):
        """Options is a list of possible value this parameter can be set to"""
        super().__init__(**kwargs)
        self.options: list = options
        self._to_repr['options'] = 'options'
        self.platypus_type = Category(self.options)

    def validate(self, value):
        return value in self.options

    def sample(self, value):
        return self.options[int(len(self.options) * value)]


class AbstractFieldSelector(Selector, ABC):
    """Base class for selectors that modify one field in one or more objects in an EnergyPlus building"""
    def __init__(self, field_name):
        super().__init__()
        self.field_name = field_name
        self._add_repr('field_name')

    def get(self, building) -> List:
        """Gets the current values of this field from a building

        :param building: the building to retrieve values from
        :return: a list containing the current values of this selector's fields
        """
        mode = ef.get_mode(building)
        objects = self.get_objects(building)
        field_name = ef.convert_format(self.field_name, 'field', mode)
        if mode == 'idf':
            return [getattr(o, field_name) for o in objects]
        if mode == 'json':
            return [o[field_name] for o in objects]

    def set(self, building, value) -> None:
        """Sets this field in the building to the provided value

        :param building: the building to modify
        :param value: the value to set this field to
        :return:
        """
        mode = ef.get_mode(building)
        objects = self.get_objects(building)
        field_name = ef.convert_format(self.field_name, 'field', mode)
        if mode == 'idf':
            for o in objects:
                setattr(o, field_name, value)
        if mode == 'json':
            for o in objects:
                assert field_name in o, f'{field_name} not in {repr(o)}'
                o[field_name] = value

    def multiply(self, building, value) -> None:
        """Multiplies this field in the building by the provided value

        :param building: the building to modify
        :param value: the value to multiply this field by
        :return:
        """
        mode = ef.get_mode(building)
        objects = self.get_objects(building)
        field_name = ef.convert_format(self.field_name, 'field', mode)
        if mode == 'idf':
            for o in objects:
                setattr(o, field_name, value * getattr(o, field_name))
        if mode == 'json':
            for o in objects:
                assert field_name in o, f'{field_name} not in {repr(o)}'
                o[field_name] = value * o[field_name]

    def add(self, building, value) -> None:
        """Adds to this field in the building by the provided value

        :param building: the building to modify
        :param value: the value to add to this field
        :return:
        """
        mode = ef.get_mode(building)
        objects = self.get_objects(building)
        field_name = ef.convert_format(self.field_name, 'field', mode)
        if mode == 'idf':
            for o in objects:
                setattr(o, field_name, value + getattr(o, field_name))
        if mode == 'json':
            for o in objects:
                assert field_name in o, f'{field_name} not in {repr(o)}'
                o[field_name] = value + o[field_name]

    @abstractmethod
    def get_objects(self, building) -> List:
        """Returns a list of the object this selector applies to

        :param building: the building to search for objects
        :return: a list of the objects this selector applies to
        """
        pass


class FilterSelector(AbstractFieldSelector):
    """A selector that uses a custom function to find which objects it should modify"""
    def __init__(self, get_objects, field_name):
        """

        :param get_objects: a function that takes a building and returns the objects this selector should modify
        :param field_name: the field to modify
        """
        super().__init__(field_name)
        self._get_objects = get_objects
        self._add_repr('get_objects', '_get_objects')

    def get_objects(self, building):
        return self._get_objects(building)


class FieldSelector(AbstractFieldSelector):
    """A selector that modifies one or more fields in an EnergyPlus building,
    based on the class, object and field names"""
    def __init__(self, class_name=None, object_name=None, field_name=None):
        """

        :param class_name: class of the object to modify
            ex: 'Material'
        :param object_name: name of the object to modify
            ex: 'Mass NonRes Wall Insulation'
        :param field_name: name of the field to modify
            ex: Thickness
        """
        super().__init__(field_name=field_name)
        self.class_name = class_name
        self.object_name = object_name
        self._add_reprs(['class_name', 'object_name'], check=True)

    def get_objects(self, building) -> List:
        """Retrieves the objects that this selector will affect from the building.

        :param building: the building to search
        :return: a list of the objects found
        """
        mode = ef.get_mode(building)
        if mode == 'idf':
            if self.class_name is not None:
                class_name = ef.convert_format(self.class_name, 'class', 'idf')
            else:
                class_name = None
            if self.object_name == '*':
                if class_name is None:
                    raise TypeError("When object_name='*', class_name must be specified.")
                return building.idfobjects[class_name]
            if self.object_name and class_name:
                # this is probably the most reliable way to select an idfObject.
                return [building.getobject(key=class_name, name=self.object_name)]
            if self.object_name:
                # There should only one object matching the name, assuming the idf is valid
                return [eppySupport.get_idfobject_from_name(building, self.object_name)]
            if class_name is not None:
                # assume that we want the first object matching the key
                # TODO: is this specific enough, or should we remove it? our JSON code does not support this
                return [building.idfobjects[class_name][0]]
            else:  # we have neither object_name nor class_name
                raise TypeError('Either class_name or object_name must be specified.')
        elif mode == 'json':
            if self.object_name == '*':
                if not self.class_name:
                    raise TypeError("When object_name='*', class_name must be specified.")
                return list(building[self.class_name].values())
            if self.object_name and self.class_name:
                # this is probably the most reliable way to select an idfObject.
                return [building[self.class_name][self.object_name]]
            if self.object_name:
                # There should only one object matching the name, assuming the building is valid
                result = [obj for objs in building.values()
                          for name, obj in objs.items()
                          if name == self.object_name]
                if len(result) != 1:
                    warnings.warn(f'found {len(result)} objects with object_name: {self.object_name}, expected 1')
                return result
            if self.class_name:
                result = list(building[self.class_name].items())
                if len(result) == 1:
                    return result
                raise ValueError(f'multiple objects with class_name {self.class_name}.'
                                 f'Cannot guarantee a reliable ordering')
            else:  # we have neither object_name nor class_name
                raise TypeError('Either class_name or object_name must be specified.')
        raise ModeError(mode)


class GenericSelector(Selector):
    """A selector that supports custom get/set functions"""

    def __init__(self, set=None, get=None, setup=None):
        super().__init__()
        self._set = set
        self._get = get
        self._setup = setup
        for attr in ('set', 'get', 'setup'):
            self._add_repr(attr, f'_{attr}', True)

    def set(self, building, value):
        if self._set is not None:
            return self._set(building, value)
        raise NotImplementedError

    def get(self, building):
        if self._get is not None:
            return self._get(building)
        raise NotImplementedError

    def setup(self, building):
        if self._setup is None:
            return
        self._setup(building)


# Parameters

class Parameter(IOBase):

    mode_options = {'Set', 'Multiply', 'Add'}

    def __init__(self, selector: Selector = None, value_descriptor: Descriptor = None, mode = 'Set', **kwargs):
        """

        :param selector: a Selector describing how to modify the building
        :param value_descriptor: a Descriptor specifying which values to use
        :param setup: a setup function to run on the building
        :param mode: the desired transformation function to use
            Set: set the sample as the value
            Multiply: multiply the pre-existing value by the sample factor
            Add: add the sample value onto the pre-existing value
        """
        super().__init__(**kwargs)
        self.selector = selector or DummySelector()
        self.value_descriptor = value_descriptor or AnyValue()
        assert mode in self.mode_options, f'Invalid mode option'
        self.mode = mode
        self._add_reprs(['selector', 'value_descriptor'], True)

    def transformation_function(self, building, value) -> None:
        """Mutates the building based on the value provided."""
        if self.mode is 'Multiply':
            self.selector.multiply(building, value)
        if self.mode is 'Add':
            self.selector.add(building, value)
        if self.mode is 'Set':
            self.selector.set(building, value)

    def sample(self, value: float):
        """Takes a value in the range 0-1 and returns a valid value for this parameter"""
        return self.value_descriptor.sample(value)

    def validate(self, value):
        """Checks if value is a valid value for this parameter.

        :param value:
        :return: True if the value is valid False otherwise
        """
        return self.value_descriptor.validate(value)

    def setup(self, building) -> None:
        self.selector.setup(building)

    @property
    def _default_name(self):
        """The name to use for this Parameter if no name was provided"""
        if hasattr(self.selector, 'field_name'):
            return self.selector.field_name

    @property
    def platypus_type(self):
        """The platypus equivalent of this parameter"""
        return self.value_descriptor.platypus_type


def wwr(value_descriptor=RangeParameter(0.01, 0.99), **kwargs) -> Parameter:
    """Makes a window-to-wall-ratio parameter.

    :param value_descriptor: a parameter describing the valid window-wall-ratios
    :return: Parameter
    """
    if isinstance(value_descriptor, RangeParameter):
        min_val, max_val = value_descriptor.min, value_descriptor.max
        if not (0 < min_val < max_val < 1):
            if 0 == min_val:
                raise ValueError('min must be strictly greater than 0')
            if 1 == max_val:
                raise ValueError('max must be strictly less than 1')
            raise ValueError('Invalid min and max values. 0 < min < max < 1 must be satisfied.')
    else:
        warnings.warn(f'wwr is intended to be used with RangeParameter. Your value_descriptor is {value_descriptor}')

    def set(building, value):
        ef.wwr_all(building, value)

    def get():
        # This feature has not yet been requested
        raise NotImplementedError('Calculation of window to wall ratio is not supported.')

    def setup(building):
        ef.one_window(building)

    selector = GenericSelector(set=set, get=get, setup=setup)
    name = kwargs.pop('name', 'Window to Wall Ratio')
    return Parameter(selector=selector, value_descriptor=value_descriptor, name=name, **kwargs)


# Simpler ways to get specific transformers
keyFormat = Dict[str, Dict[str, Tuple[float, float]]]


# TODO: Formalize this syntax, maybe add a guaranteed order
def expand_plist(pList: keyFormat) -> List:
    """This function expands a nested dictionary of the correct format into a list of inputs.

    The dictionary should have the format:
    {'idf object name': {'idf object1 property name': (min_value, max_value)}

    Both layers of the dictionaries can have as many names as desired"""
    return [Parameter(FieldSelector(object_name=name, field_name=prop),
                      RangeParameter(min_val=min_, max_val=max_), name=prop)
            for name, subProps in pList.items()
            for prop, (min_, max_) in subProps.items()]


if __name__ == '__main__':
    pass
    # a few simple tests of the inputs module
    #edit: moved tests to test_parameter.py

    #import eppy_funcs as ef

    # transformation functions should modify the idf
    # TODO: Test transformation functions

    #print('tests passed')



from typing import List, Dict, Tuple, Union
from dataclasses import dataclass
from abc import ABC, abstractmethod

import os
import re
from itertools import chain
import pandas as pd

import config
from errors import ModeError
from IO_Objects import Objective
import eppy_funcs as ef


def get_data_dict_regex():
    """Generates a regex expression that can parse a line of the data dictionary in the .mtr file
    Various assumptions are made about the exact composition of names."""
    report_code = r'(?P<report_code>\d+)'  # report code of digits
    num_items = r'(?P<num_items>\d+)'  # number of items composed of digits
    name = r'(?P<name>[^\[\]]+?)'  # assumes that name can be anything without '[]' characters, may be wrong
    units = r'\[(?P<units>[^\]]*?)\]'  # units are enclosed in square brackets.
    frequency = r'!(?P<frequency>[\w|,]+)'  # frequency preceded by a '!'
    remainder = r'(?:  ?\[(?P<remainder>[^\]]+)\])?'  # can be preceded by 1-2 spaces based on experience
    data_dict_line = re.compile(fr'^{report_code},{num_items},{name} {units} {frequency}{remainder}$')
    return data_dict_line


_values_collection = List[List[float]]


@dataclass
class EPResults:
    """Represents the information about a Meter read from a .mtr file."""
    name: str
    frequency: str
    units: str
    data: pd.DataFrame

    # TODO: Consider improving the index for the DataFrame
    # index derived from the section headers (i.e. Design day winter hour=1) converted to a timestamp
    # this would simplify manipulating time series
    # Also consider nicer handling of duplicate names than just numbering

    @classmethod
    def from_pieces(cls, name, frequency, units, items, data, prefix=None) -> 'EPResults':
        if prefix is True:
            prefix = name
        unq_names = {}
        for i, original in enumerate(items):
            unq_names[original] = unq_names.get(original, []) + [i]
        new_names = items[:]
        for name, values in unq_names.items():
            if len(values) == 1:
                continue  # not a duplicate, leave the name alone
            for count, index in enumerate(values):
                new_names[index] = f'{name}_{count}'
        if prefix is not None:
            names = [f'{prefix}_{name}' for name in new_names]
        else:
            names = new_names
        df = pd.DataFrame({name: values for name, values in zip(names, zip(*data))})
        return cls(name, frequency, units, df)


_results_format = Dict[Tuple[str, str], EPResults]


def read_eso(out_dir: str = config.out_dir, file_name="eplusout.eso") -> _results_format:
    """Retrieve information from an Energy-Plus simulation, with outputs in out_dir

    returns a dictionary with keys of the form {(MeterName, Reporting_Frequency or None):EPResults dataclass}"""
    # output format described here:
    # https://energyplus.net/sites/default/files/pdfs_v8.3.0/OutputDetailsAndExamples.pdf

    with open(os.path.join(out_dir, file_name), 'r') as file:
        f = file.readlines()  # may cause memory errors on very large files

    # TODO: Find a way to avoid hardcoding these values to make things portable between versions
    # or move them to a version-dependant config of some kind
    data_dictionary_start = 1  # omit the first line, it is E+ version info
    data_dictionary_header = 6  # header is always 5 lines for E+ 8.8 and 6 lines for E+ 9.0
    data_dictionary_end = f.index('End of Data Dictionary\n')
    data_end = -2  # omit the End of Data line and the Number of Records Written line

    header = f[data_dictionary_start:data_dictionary_header]  # header currently ignored
    requested_vars = f[data_dictionary_header + 1: data_dictionary_end]
    data = f[data_dictionary_end + 1:data_end]

    data_dict_line = get_data_dict_regex()
    codes = {}
    keys = set()
    # currently ignores the header
    for line in requested_vars:
        match = re.search(data_dict_line, line)
        if match is None:
            raise ValueError(f'could not match: {line}')
        name = match.group('name')
        frequency = match.group('frequency')
        items = match.group('remainder')
        items = ['Value'] if items is None else items.split(',')
        report_code = match.group('report_code')
        key = (name, frequency)
        assert report_code not in codes, f'already found report code {report_code}'
        assert key not in keys, f'already found item with {key}'

        result = dict(name=name, frequency=frequency,
                      units=match.group('units'),
                      items=items, data=[])

        codes[report_code] = result
        keys.add(key)
    for line in data:
        report_code, *values = line.split(',')
        if report_code in codes:
            values = [float(value) for value in values]
            codes[report_code]['data'].append(values)

    return {(result['name'], result['frequency']): EPResults.from_pieces(**result) for result in codes.values()}


def sum_values(results: EPResults) -> float:
    """Returns the sum over the Value column from some results

    :param values: the values to sum over
    :return: the sum of the first value for each entry in the collection.
    """
    return results.data['Value'].sum()


# TODO make "all" clear all optional output, currently incomplete (i.e. csv files)
def clear_outputs(building, outputs: Union[str, List[str]] = 'all') -> None:
    """Disable certain types of output.

    :param building: the building to modify
    :param outputs: Can be the class_name of the output to clear,
        a shortcut for different types of output,
        or a list that combines the above.
    :return: None
    """
    # `empty` is used so that different fields do not reference each the same container
    mode = ef.get_mode(building)
    if mode == 'idf':
        building_items = building.idfobjects
        empty = list
    elif mode == 'json':
        building_items = building
        empty = dict
    else:
        raise ModeError(mode)

    def find(prefix):
        return {x for x in building_items if x.startswith(ef.convert_format(prefix, 'class', mode))}

    # These lists are probably incomplete!
    class_names = dict()
    # ENVIRONMENTALIMPACTFACTORS adds a bunch of meters, but is not a meter itself
    class_names['output_meters'] = find('Output:Meter:') | {'Output:Meter', 'Output:EnvironmentalImpactFactors'}
    class_names['internal_meters'] = {'Meter:Custom', 'Meter:CustomDecrement'}
    class_names['tables'] = find('Output:Table:')

    class_names['output'] = find('Output:')
    class_names['outputcontrol'] = find('OutputControl:')

    class_names['meters'] = class_names['output_meters'] | class_names['internal_meters']

    if outputs == 'all':
        outputs = chain(*class_names.values())
    else:
        if isinstance(outputs, str):  # make outputs a list
            outputs = [outputs]
        # convert each element in outputs via the class_names shortcuts
        outputs = set(chain(*(class_names.get(output, [output]) for output in outputs)))
    # clear the selected outputs
    for output in outputs:
        building_items[output] = empty()


class EPReader(Objective, ABC):
    field_pairs = NotImplemented

    def __init__(self, class_name, frequency: str = None, func=sum_values, **kwargs):
        super().__init__(**kwargs)
        self.class_name = class_name
        self.frequency = frequency
        # class name has a nonzero default, so this may be adding noise to the repr
        # but it does provide a more complete picture
        self._add_reprs(['class_name', 'frequency'], check=True)

        self._process = func
        self._add_repr('func', '_process')

    def check_all(self, objective, mode):
        if mode == 'idf':
            def get(attribute):
                return getattr(objective, attribute)
        elif mode == 'json':
            def get(attribute):
                return objective[attribute]
        else:
            raise ModeError(mode)
        for self_attr, objective_attr in self.field_pairs:
            self_value = getattr(self, self_attr)
            objective_value = get(ef.convert_format(objective_attr, 'field', mode))
            if self_value is not None and objective_value != self_value:
                return False
        return True

    def get_objective(self, building):
        mode = ef.get_mode(building)
        if mode == 'idf':
            objectives = building.idfobjects[ef.convert_format(self.class_name, 'class', mode)]
        elif mode == 'json':
            objectives = building[ef.convert_format(self.class_name, 'class', mode)].values()
        else:
            raise ModeError(mode)
        for objective in objectives:
            if self.check_all(objective, mode):
                return objective
        raise ValueError(f'Cannot find the objective for {repr(self)}')

    def add_objective(self, building):
        """Creates and adds the meter needed by this objective to the building.

        :param building: the building to modify
        :return: None
        """
        try:
            self.get_objective(building)
        except ValueError:
            pass  # the objective is not present
        else:
            raise ValueError(f'Objective for {repr(self)} already exists')

        mode = ef.get_mode(building)
        new_object_dict = {ef.convert_format(field, 'field', mode): getattr(self, attr)
                           for attr, field in self.field_pairs if getattr(self, attr)}
        if self.frequency is None:
            new_object_dict[ef.convert_format('Reporting_Frequency', 'field', mode)] = 'Hourly'
        if mode == 'idf':
            building.newidfobject(key=ef.convert_format(self.class_name, 'class', mode), **new_object_dict)
        elif mode == 'json':
            # this is equivalent to appending, but e+ uses a dictionary instead of a list
            objectives = building[self.class_name]
            num = len(objectives)
            new_key = f'{self.class_name} {num}'
            assert new_key not in building, f'The building has incorrectly numbered {self.class_name} entries'
            objectives[new_key] = new_object_dict

    def validate(self, building):
        self.get_objective(building)

    def setup(self, building) -> None:
        """Prepares an idf so that it's output can be read by this meter.

        :param building: the idf to modify
        :return: None
        """
        try:
            self.get_objective(building)
        except ValueError:
            self.add_objective(building)
        self.validate(building)

    @abstractmethod
    def results_name(self):
        pass

    def __call__(self, results: _results_format) -> float:
        results_name = self.results_name()
        if self.frequency:
            meter_results = results[(results_name, self.frequency)]
        else:
            for (name, _), v in results.items():
                if name == results_name:
                    meter_results = v
                    break
            else:
                raise ValueError(f'No meter with name {results_name} found')
        return self._process(meter_results)


class MeterReader(EPReader):
    field_pairs = (('key_name', 'Key_Name'), ('frequency', 'Reporting_Frequency'))

    def __init__(self, key_name, class_name=config.objective_meter_type,
                 frequency: str = None, func=sum_values, **kwargs):
        super().__init__(class_name=class_name, frequency=frequency,
                         func=func, **kwargs)
        self.key_name = key_name
        self._add_repr('key_name', check=True)

    def results_name(self):
        return self.key_name

    @property
    def _default_name(self):
        return self.key_name


class VariableReader(EPReader):
    field_pairs = (('key_value', 'Key_Value'),
                   ('variable_name', 'Variable_Name'),
                   ('frequency', 'Reporting_Frequency'))

    def __init__(self, key_value, variable_name='*', class_name=config.objective_variable_type,
                 frequency: str = None, func=sum_values, **kwargs):
        super().__init__(class_name=class_name, frequency=frequency, func=func, **kwargs)
        self.key_value = key_value
        self._add_repr('key_value')
        self.variable_name = variable_name
        if self.variable_name != '*':
            self._add_repr('variable_name')

    def results_name(self):
        return f'{self.key_value},{self.variable_name}'

    @property
    def _default_name(self):
        return self.variable_name

# TODO: Remove objectives from idf output automatically



import pytest

import eppy_funcs as ef
from evaluator import EvaluatorEP, EvaluatorSR
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector
from problem import EPProblem, Problem
import sampling


@pytest.fixture
def building():
    #returns the basic building
    return ef.get_building()


@pytest.fixture
def problem():
    parameters = [Parameter(FieldSelector(object_name='Mass NonRes Wall Insulation', field_name='Thickness'))]
    objectives = ['Electricity:Facility', 'Gas:Facility'] # the default is just 'Electricity:Facility'

    problem=EPProblem(parameters, objectives) #EPP Problem automatically converts these to MeterReaders
    return problem


def test_evaluatorEP(building, problem):
    """To make sure EvaluatorEP can be initialised and works as intended"""

    evaluator = EvaluatorEP(problem, building)
    result = evaluator([0.5]) # run with thickness set to 0.5

    assert result == (1818735943.9307632, 2172045529.871896), f'Unexpected result for EvaluatorEP, {result}'
    #change this to 0 to see stdout and stderr
    assert 1


def test_evaluatorSR(building, problem):
    """To make sure EvaluatorSR can be initialised and works as intended"""
    def function(values):
        return ((values[0], values[0]**2), ())

    # this denotes a problem which takes 1 input, produces 2 outputs and no constraints. The placeholder parameters/objectives will be generated automatically.
    new_problem = Problem(1,2,0)

    evaluator_1 = EvaluatorSR(function, problem)
    evaluator_2 = EvaluatorSR(function, new_problem)
    result_1 = evaluator_1([4])
    result_2 = evaluator_2([4])

    assert result_1 == (4, 16), f'Unexpected result for EvaluatorSR with EPProblem, {result_1}'
    assert result_2 == (4, 16), f'Unexpected result for EvaluatorSR with custom problem, {result_2}'

    #change this to 0 to see stdout and stderr
    assert 1



import os
from eppy import modeleditor
from eppy.modeleditor import IDF
from eppy.results import readhtml
import inspect
import subprocess

def listify_idf_objects(idf: IDF, objects: list):
	"""	listify_idf_objects(IDF, list)

		Takes the passed idf, and grabs the objects out of it
		that are specified in objects

		idf(IDF): The idf to grab the objects from
		objects(list): A list generated from a yaml file
			       that specifies how and what to read
			       from the idf file
	"""
	objects_list = []

	#: For every set of object information in objects, and for
	#: every attribute in that object, check if it's in the variables
	#: dictionary, and if it is, and its a float, then append it to
	#: objects_list
	for objectset in objects:

		#: If the name of the object is "Run_Parameters", then its
		#: not an object, its actually something in the config file
		#: used to setup some run inputs.
		if objectset[0] == "Run_Parameters":
			continue

		#: Get the idf object from the idf based off its name (objectset[0])
		current_object = get_idfobject_from_name(idf, objectset[0])

		#: Loop through the attributes and check if both the attribute
		#: is in the Variables dictionary, and that it's float
		for attribute in current_object.fieldnames:
			#: Catch value errors to check if attributes are floats.
			try:
				#: Check that the attribute is in the Variables dictionary,
				#: if its not, raise a ValueError to move on.
				if attribute not in objectset[1]["Variables"]:
					raise ValueError

				#: Store a float of the attribute (if it cannot
				#: be converted to a float, then it will raise a
				#: ValueError, and move on.
				val = float(getattr(current_object, attribute))

				#: Append the value to the objects_list
				objects_list.append(val)

			#: Except the ValueError, and pass
			except ValueError as e:
				pass

	#: Return the list of objects
	return objects_list

def rebuild_idf(idd_path: str, idf_path: str, epw_path: str, w_objects: list, build_list: list):
	"""	rebuild_idf(IDF, list, list)

		Builds an idf based off of a passed config file, and
		a list of values that was passed.

		Args:
			idd_path(str):	The absolute path to the idd file to be run.

			idf_path(str):	The absolute path to the idf file to be run.

			epw_path(str):	The absolute path to the epw file to be run.

			w_objects(list):	A list generated from a yaml file
						that specifies how and what to read
						from the idf file

			build_list(list):	A list of values which will be assigned to
						values in objects in the idf.
	"""

	#: Get an idf object from the passed idd, idf, and epw
	idf = create_idf_object(idd_path, idf_path, epw_path)
	i = 0

	#: For every set of object information in w_objects,
	#: and for every attribute in the object described if that attribute
	#: is in the Variables dictionary, and the attribute is a float, AND the
	#: value for this entry in build_list is between the max and min values
	#: described in the object's information then set the attribute to the
	#: value in build_list. (Note, the first object's first valid field is
	#: equivellant to the first item in build_list, the second field to the
	#: second item in build_list, etc)
	for object in w_objects:

		#: If the name of the object is "Run_Parameters", then
		#: this is not an idf object, it's an informatic field
		#: from the config, so ignore it.
		if object[0] == "Run_Parameters":
			continue

		#: Get the object from the idf based off its name (object[0] holds
		#: the name of the object)
		current_object = get_idfobject_from_name(idf, object[0])

		#: For each attribute in the object, check if the attribute is a
		#: float, and check that the value that the attribute that is to be
		#: set to is within the range given, and if both are true, then set
		#: the attribute to that value
		for attribute in current_object.fieldnames:

			#: Try catch block to catch ValueErrors that float(getattr) raises
			#: if the attribute is not a float (to check if the attribute is float)
			try:
				#: Check if the attribute is in the Variables dictionary
				#: and if its not, raise a ValueError to skip this attribute.
				if attribute not in object[1]["Variables"]:
					raise ValueError

				#: Attempt to get the attribute to check if its a float or not.
				val = float(getattr(current_object, attribute))

				#: If we should check max/mins
				if dict(w_objects)["Run_Parameters"][1]:

					#: If the attribute has specific max/mins,
					#: then check if its within those limits
					if attribute in object[1].keys():

						#: If the value is less than the min
						if build_list[i] < float(object[1][attribute][1]):

							#: return None as this idf is invalid
							return None

						#: If the value is more than the max
						elif build_list[i] > float(object[1][attribute][0]):

							#: return None as this idf is invalid
							return None

					#: Otherwise check it against the defaults
					else:

						#: If the value is less than the default min
						if build_list[i] < float(object[1]["Default"][1]):

							#: return None as this idf is invalid
							return None

						#: If the value is more than the default max
						elif build_list[i] > float(object[1]["Default"][0]):

							#: return None as this idf is invalid
							return None

				#: If we get to this point, then the attribute is valid, and the
				#: value is valid, so set the attribute to that value
				setattr(current_object, attribute, round(build_list[i], 2))

				#: Add 1 to i, so that we can assign the next value to the next
				#: attribute
				i += 1

			#: Except ValueErrors incase the attribute is not a float, or it is not in Variables
			except ValueError as e:
				for key, value in object[1]["Abstracts"].items():
					if attribute == key:
						print(attribute, " ==> ", value)
						setattr(current_object, attribute, round(value, 2))

	#: Return the idf with all the new attributes.
	return idf

def get_idfobject_from_name(idf: IDF, name: str):
	"""	get_idfobject_from_name(IDF, str)

		Gets an object from the passed idf where it's name
		value is equal to the passed string, if none are found
		then this method returns None

		Args:
			idf(IDF): The idf to find the object from.
			name(str): The string to find that is equal to the
				   name field of the object.
	"""

	#: For every object in the idf.
	for idfobject in idf.idfobjects:

		#: For field in this object
		for object in idf.idfobjects[idfobject]:

			#: Try except block incase object does not
			#: have a .Name attribute
			try:
				#: If object.Name == name, then return
				#: the object, as that is the object
				#: we're looking at.
				if object.Name == name:
					return object

			#: Move on to the next object if an exception is
			#: raised.
			except:
				pass

	#: If this is enocuntered then no match was found, so we return None.
	return None

def create_idf_object(idd: str, idf: str, epw: str):
	"""	create_idf_object(str, str, str)

		Returns an idf object generated from the passed
		idd, idf, and epw

		Args:
			idd(str): The directory to the idd
			idf(str): The directory to the idf
			epw(str): The directory to the epw
	"""

	#: Try to set the idd attribute for the IDF class.
	try:
		IDF.setiddname(idd)
	except:
		pass

	#: Create an idf object based off the passed idf and epw.
	idf = IDF(idf, epw)

	#: return the created idf object
	return idf

def printErr(msg: str):
	""" 	printErr(str)

		Prints the message in an error red on unicode supported
		consoles.

		Args:
			msg(str): The string to print
	"""
	#: print(red_colour + msg + end_colour)
	print('\033[91m' + msg + '\033[0m')

def printSuccess(msg: str):
	""" 	printSuccess(str)

		Prints the message in a success green on unicode supported
		consoles.

		Args:
			msg(str): The string to print
	"""
	#: print(green_colour + msg + end_colour)
	print('\033[92m'  + msg + '\033[0m')

def get_energy_usage(idf: IDF, obj_list= None) -> float:
	"""	get_energy_usage(IDF, list)

		Runs an idf, and extracts the building's energy usage
		from that run.

		Args:
			idf(IDF): The idf to be run.
			obj_list(list): ONLY USED IF VERBOSE IS TRUE FOR
					DEBUGGING PURPOSES. This variable
					prints the final state of each of
					the idf's objects that are being
					changed.
	"""

	#: Verbose is used to help debug, when true it prints
	#: additional information about each run.
	verbose = False

	#: Try to run the idf. If an error is caught during this execution
	#: then the idf is unrunable, and we will return an energy usage of inf
	try:

		#: Run the idf. if Verbose is False, then we log that we're running
		#: the idf, and when it's finished, but we suppress the output from
		#: the idf's actual run. If verbose is True, we instead run the idf
		#: without suppressing its objectives
		if not verbose:
			print("Running idf...")
			idf.run(verbose='q') #: verbose='q' suppresses objectives
			printSuccess("Done.")
		else:
			idf.run()
	except:
		#: If an exception is raised, then print that we failed to run
		#: the idf, and return inf. NOTE, if verbose is true, we will also
		#: print all of the objects being looked at, so that we can analize
		#: why the idf is unrunable
		printErr("Failed to run idf. Is it valid?")
		if verbose == True and obj_list is not None:
			for object in obj_list:
				print(get_idfobject_from_name(idf, object[0]))

		#: return inf
		return float("inf")

	#: Create a variable to hold the sum of the energy used.
	energy_used = 0.0

	#: Open the meter file to read what the energy usages were.
	with open("eplusout.mtr", 'r') as filestring:

		#: Read all the lines in the file, and split them into seperate string
		#: in a list. This way we can analize each line seperately.
		strings = filestring.read().split('\n')
		for line in strings[35:]:

			#: If the line has the id 13 (ie the first entry is 13) then
			#: this is one of the objectives we want to look at, so grab the
			#: float value from the line (which will be in the second
			#: position of the csv) and read it as a float. Add that float
			#: to the total energy used.
			if line[:3] == "13,":
				energy_used += float(line[3:line[3:].find(',')])

			#: If the line has the id 274 (ie the first entry is 274) then
			#: this is one of the objectives we want to look at, so grab the
			#: float value from the line (which will be in the second
			#: position of the csv) and read it as a float. Add that float
			#: to the total energy used.
			elif line[:4] == "274,":
				energy_used += float(line[4:line[4:].find(',')])

	#: Return the total amount of energy used.
	return energy_used



def main():
	""" 	main()

		This method will not be documented as it is not meant to be run. This method is used
		for debugging and testing only, and thus it may change from commit to commit without
		any mention.
	"""
	IDF.setiddname("EP.idd")
	idf = IDF("input.idf", "t.epw")
	myWindow = idf.idfobjects['FENESTRATIONSURFACE:DETAILED']
	for window in myWindow:
		print(window)
	exit()

	print(get_idfobject_from_name(idf, "Perimeter_ZN_3_wall_north_Window_6"))

	energies_used = []
	min_height = 0.91
	denom = 100

	for i in range(200, -1, -1):
		idf.idfobjects['FENESTRATIONSURFACE:DETAILED'][0].Vertex_1_Zcoordinate = round(min_height + (i / denom), 2)
		idf.idfobjects['FENESTRATIONSURFACE:DETAILED'][0].Vertex_4_Zcoordinate = round(min_height + (i / denom), 2)

		print("Running with window height {}".format(round(min_height + (i/denom), 2)))
		energies_used.append((round(min_height + (i / denom), 2), get_energy_usage(idf)))

	energies_used = sorted(energies_used, key= lambda x: x[1])

	for index, use in enumerate(energies_used):
		print("{}th  best solution has window height {}".format(index + 1, use[0]), end='\n\t')
		print(use[1])

	print("Best window size is: ", end='\n\t')
	print(min(energies_used, key=lambda x: x[1]))


if __name__ == "__main__":
	main()



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

from besos.parameters import CategoryParameter, RangeParameter, Parameter
from besos.evaluator import EvaluatorSR
from besos.problem import Problem
from besos import optimizer
from platypus.config import PlatypusConfig

# Optimizing a building usually requires many evaluations of that building at different parameters. 
# Therefore this example will optimize over a surrogate evaluator wrapping a quick function 
# to speed things up.  
# By default, platypus does not support using categories and ranges in the same problem. The wrapped algorithms  
# available in the optimizer package *do* support mixed type optimization.

# +
parameters=[Parameter(value_descriptor=CategoryParameter(list(range(80,90))), name='1st'),
            Parameter(value_descriptor=CategoryParameter(list(range(80,90))), name='2nd'),
            Parameter(value_descriptor=RangeParameter(3, 100), name='3rd')
           ]

def mixed_types(vals):
    vals = list(vals)
    num = vals.pop(-1)
    objectives = tuple(num % v for v in vals)
    return (objectives,())


evaluator = EvaluatorSR(mixed_types, Problem(parameters, 2))
# -

# by default variator='automatic' i.e. it will be constructed based on the evaluator used
optimizer.EpsMOEA(evaluator, epsilons=10)

# If you are customizing more of your optimization, you can acces the mixed type operator for your evauluator
# using the function `optimizer.get_operator` (Use mutation=True if you want a mutator instead of a variator)

# +
variator = optimizer.get_operator(evaluator.to_platypus())

result = optimizer.EpsMOEA(evaluator, variator=variator, epsilons=10)
result
# -





import pytest

import eppy_funcs as ef
from evaluator import EvaluatorEP, EvaluatorSR
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector, wwr, expand_plist
from problem import EPProblem, Problem
from optimizer import NSGAII
from objectives import MeterReader, VariableReader, clear_outputs
import sampling


@pytest.fixture
def building():
    #returns the basic building
    building = ef.get_building()
    return building


@pytest.fixture
def parameters():
    #The parameters for the problem/evaluator
    parameters=expand_plist(
        {'NonRes Fixed Assembly Window':
            {'U-Factor':(0.1,5), 'Solar Heat Gain Coefficient':(0.01,0.99)
        },
        'Mass NonRes Wall Insulation':
            {'Thickness':(0.01,0.09)},
        })
    return parameters


def test_objectives(building, parameters):
    """Testing custom functions and basic objective creation"""

    def variance(result):
        return result.data['Value'].var()

    objectives = [MeterReader('Electricity:Facility', name='Electricity Usage'),
              MeterReader('Electricity:Facility',func=variance, name='Electricity Variance')]
    problem = EPProblem(inputs=parameters, outputs=objectives)

    evaluator = EvaluatorEP(problem, building)
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, 10)
    results = evaluator.df_apply(samples, keep_input=True)

    value = results.iloc[0]['Electricity Variance']

    assert value == 829057663033101.1, f'Unexpected value when using custom function:{value}'
    #change this to 0 to see stdout and stderr
    assert 1


def test_constraints(building, parameters):
    """Testing for expected output with certain constraints, also acts as a test for using NSGAII"""

    objectives = ['Electricity:Facility', 'Gas:Facility']
    problem = EPProblem(inputs=parameters,
                        outputs=objectives,
                        constraints=['CO2:Facility'], constraint_bounds=['>=750'])

    evaluator = EvaluatorEP(problem, building)

    results = NSGAII(evaluator, evaluations=10, population_size=2)

    value = results.iloc[0]['CO2:Facility'] + results.iloc[0]['violation']
    assert value >= 750, f'Constraint did not effect output, value should be above 750 but was: {value}'

    #Check to make sure the output changes with different constraints
    problem = EPProblem(inputs=parameters,
                        outputs=objectives,
                        constraints=['CO2:Facility'], constraint_bounds=['<=750'])

    evaluator = EvaluatorEP(problem, building)

    results = NSGAII(evaluator, evaluations=10, population_size=2)

    value = results.iloc[0]['CO2:Facility'] - results.iloc[0]['violation']
    assert value <= 750, f'Constraint did not effect output, value should be below 750 but was: {value}'

    #change this to 0 to see stdout and stderr
    assert 1



import pytest

from optimizer import rbf_opt
from evaluator import EvaluatorSR
from parameters import Parameter, RangeParameter
from problem import Problem

def test_rbfopt():
    """Simple test to make sure RBFopt can be run properly"""

    def obj_funct(x):
        return (x[0]*x[1] - x[2],),()

    param_list = [Parameter(value_descriptor=RangeParameter(0,10)) for _ in range(3)]
    problem = Problem(param_list, 1)

    evaluator = EvaluatorSR(obj_funct, problem, error_mode='Silent')

    opt = rbf_opt(evaluator, 30)
    value = opt.iloc[0]

    assert value['None_0'] == 0 and value['None_1'] == 5.062882681226724 and value['None_2'] == 10 and value['outputs_0'] == -10, f'Unexpected output: {value}'



import pytest
import pandas as pd
from pandas import DataFrame as DF
import numpy as np

from sklearn import svm, pipeline, linear_model
from sklearn.preprocessing import StandardScaler

import eppy_funcs as ef
from evaluator import EvaluatorEP, EvaluatorSR
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector
from problem import EPProblem
import sampling

def test_factor():
    building = ef.get_building()

    parameters = [
                Parameter(FieldSelector('Lights', '*', 'Watts per Zone Floor Area'),
                    value_descriptor=CategoryParameter([1,2,3,4,5,6,7,8,9,10]),
                    name='Lights Watts/Area', mode='Add')]
    objectives = ['Electricity:Facility']
    problem = EPProblem(parameters, objectives)
    evaluator = EvaluatorEP(problem, building, multi = False)

    parameter2 = Parameter(FieldSelector('Lights', '*', 'Watts per Zone Floor Area'),
                    value_descriptor=RangeParameter(.5,2),
                    name='Lights Watts/Area', mode='Multiply')
    parameters2 = [parameter2]
    objectives2 = ['Electricity:Facility']
    problem2 = EPProblem(parameters2, objectives2)
    evaluator2 = EvaluatorEP(problem2, building, multi = False)

    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, 20)
    samples2 = sampling.dist_sampler(sampling.seeded_sampler, problem2, 20)

    outputs = evaluator.df_apply(samples, keep_input = True, processes = 4)
    outputs2 = evaluator2.df_apply(samples2, keep_input = True, processes = 4)

    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)

    print()
    print()
    print(outputs)
    print(problem.names('inputs'))
    print(problem.inputs[0].mode)
    print(type(problem.inputs))
    print()
    print()
    print(outputs2)
    #change this to 0 to see stdout and stderr
    assert 1



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorEP, AdaptiveSR
from besos.parameters import RangeParameter, expand_plist, wwr
from besos.problem import EPProblem

import pyKriging
from pyKriging.krige import kriging
import numpy as np

import pandas as pd

from typing import Union, List, Tuple
# -

# Adaptive Surrogate models require more setup that other parts of BESOS, since they have several optional features, and there are different approaches to using them. Below is the template that can be used to make an adaptive surrogate.

# +
tabular = Union[np.array, pd.DataFrame]
DF = pd.DataFrame

class ExampleTemplate(AdaptiveSR):
    # helper functions provided by AdaptiveSR (Generally avoid editing these, but use them as needed)
    # append_data(X, y)
    # do_infill
    # get_from_reference
    
    # functions with defaults (These can be removed from this template if you like the defaults)
        # They may depend on some of the optional functions in order to work if using the defaults
    # __init__
    # infill (depends on -> get_infill)
    # update_model
    
    # optional functions (These will not work unless you implement them)
    # get_infill
    
    # required functions
    # train
    # eval_single
    
    def __init__(self, custom_argument_1, custom_arg_2, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # do stuff with custom arguments here
    
    def get_infill(self, num_datapoints: int) -> tabular:
        """Generates data that is most likely to improve the model, and can be used for retraining.

        :param num_datapoints: the number of datapoints to generate
        :return: the datapoints generated, in some tabular datastructure
        """
        raise NotImplementedError

    def update_model(self, new_data: tabular, old_data: DF = None) -> None:
        """Modifies self.model to incorporate the new data.

        This function should not edit the existing data

        :param new_data: a table of inputs and outputs
        :param old_data: the table of inputs and outputs without the new data
        :return: None
        """
        self.train()
        
    def infill(self, num_datapoints: int) -> None:
        """Adds num_datapoints samples to the model and updates it.

        :param num_datapoints: number of datapoints to add to the model's training set
        :return: None
        """
        inputs: DF = self.problem.to_df(self.get_infill(num_datapoints), 'inputs')
        outputs: DF = self.get_from_reference(inputs)
        self.do_infill(pd.concat([inputs, outputs], axis=1))

    def train(self) -> None:
        """Generates a new model using the stored data, and stores it as self.model"""
        pass

    def eval_single(self, values: List) -> Tuple:
        """Evaluates a single input point

        :param values: The datapoint to evaluate
        :param **kwargs:
        :return: A tuple of the predicted outputs for this datapoint
        """
        pass
# -

# Here we will go through piece by piece how to wrap a model from pyKrigging in our setup.

class KrigingEval(AdaptiveSR):
    # The model cannot handle multiple objectives, so we check for this when initializing
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.problem.num_outputs != 1:
            raise ValueError('This model cannont handle multiobjective problems.')
        if self.problem.num_constraints != 0:
            raise ValueError('This model cannont handle constrained problems.')
    
    # pyKriging models are able to generate datapoints which will improve the model the most
    # **kwargs is included, since the model's function has some options which we may want to access
    # such as the ability to sample near the minimum, instead of near the point of greatest uncertainty
    def get_infill(self, num_datapoints, **kwargs):
        return self.model.infill(num_datapoints, **kwargs)
    
    # since pyKriging's model has an addPoint method, we will use it to update the model instead
    # of initialising a new one each time. We loop through all of the new datapoints
    # and add them one at a time, then retrain the inner model.
    # note that self.model.train is not the same as self.train
    def update_model(self, new_data, old_data = None) -> None:
        for index, *row in new_data.itertuples():
            inputs = row[:self.problem.num_inputs]
            output = row[-1]
            assert len(row) == self.problem.num_inputs + self.problem.num_outputs
            self.model.addPoint(inputs, output)
        self.model.train()
    
    
    
    # The infill function will work automatically, since we have defined get_infill and update_model
    
    
    # we initialize and store a kriging model on the stored training data, and then
    # run the internal model's train function.
    def train(self):
        self.model = kriging(self.data.values[:,:-1], self.data.values[:,-1])
        self.model.train()
    
    # this model expects a 2d array representing a batch of inputs, but we only want 
    # to evaluate one input point at a time, so we wrap the inputs in a list before passing
    # them to the model
    def eval_single(self, values):
        return ((self.model.predict(list(values)),), ())

# +
building = ef.get_building()

parameters=expand_plist(
    {'Mass NonRes Wall Insulation':
      {'Thickness': (0.05, 0.99)}
    })

parameters.append(wwr())

problem = EPProblem(parameters)

evaluator = EvaluatorEP(problem, building)

inputs = sampling.dist_sampler(sampling.lhs, problem, 10)
inputs = sampling.add_extremes(inputs, problem)
# this helps this specific evaluator auto-detect the min/max correctly
# -

x = sampling.dist_sampler(sampling.lhs, problem, 10)
x

sampling.add_extremes(x, problem)

k = KrigingEval(reference=evaluator)

k.do_infill(inputs) # even with cached evaluator results, this will take time for the pyKriging train step

k.model.train()

k.model.plot()

k.infill(5) 

k.model.plot()

k.data





import pytest

import eppy_funcs as ef
from evaluator import EvaluatorEP, EvaluatorSR
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector
from problem import EPProblem, Problem
import sampling


@pytest.fixture
def building():
    return ef.get_building()


@pytest.fixture
def parameters():
    parameters = [
        Parameter(FieldSelector(object_name='NonRes Fixed Assembly Window',
                field_name='Solar Heat Gain Coefficient'),
                value_descriptor=RangeParameter(0.01,0.99)),
        Parameter(FieldSelector('Lights', '*', 'Watts per Zone Floor Area'),
                value_descriptor=RangeParameter(8, 12),
                name='Lights Watts/Area')]
    return parameters


@pytest.fixture
def problem(parameters):
    objectives = ['Electricity:Facility']
    problem = EPProblem(parameters, objectives)
    return problem


def test_eval(building, problem):
    evaluator = EvaluatorEP(problem, building, multi=True)
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, 200)

    outputs = evaluator.df_apply(samples, keep_input = True, processes = 4)
    print()
    print(outputs)
    evaluator.estimate_time(samples, processes = 4)
    assert 1



import setuptools

with open("besos/README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="besos",
    version="1.0.3",
    description="A library for Building and Energy Simulation, Optimization and Surrogate-modelling",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude="pytests"),
    include_package_data=True,
    install_requires=[
        'eppy',
        'pyDOE2',
        'numpy',
        'pandas',
        'platypus-opt',
        'rbfopt',
        'matplotlib',
        'ipywidgets',
        'jupytext',
        'pathos',
        'sklearn',
        'pyKriging',
        'pyehub',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



import pytest

import eppy_funcs as ef
import sampling
from evaluator import EvaluatorEP, AdaptiveSR
from parameters import RangeParameter, expand_plist, wwr
from problem import EPProblem

import pyKriging
from pyKriging.krige import kriging
import numpy as np

import pandas as pd

from typing import Union, List, Tuple


#Class taken directly from the notebook
class KirgingEval(AdaptiveSR):
    # The model cannot handle multiple objectives, so we check for this when initializing
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.problem.num_outputs != 1:
            raise ValueError('This model cannont handle multiobjective problems.')
        if self.problem.num_constraints != 0:
            raise ValueError('This model cannont handle constrained problems.')

    # pyKriging models are able to generate datapoints which will improve the model the most
    # **kwargs is included, since the model's function has some options which we may want to access
    # such as the ability to sample near the minimum, instead of near the point of greatest uncertainty
    def get_infill(self, num_datapoints, **kwargs):
        return self.model.infill(num_datapoints, **kwargs)

    # since pyKriging's model has an addPoint method, we will use it to update the model instead
    # of initialising a new one each time. We loop through all of the new datapoints
    # and add them one at a time, then retrain the inner model.
    # note that self.model.train is not the same as self.train
    def update_model(self, new_data, old_data = None) -> None:
        for index, *row in new_data.itertuples():
            inputs = row[:self.problem.num_inputs]
            output = row[-1]
            assert len(row) == self.problem.num_inputs + self.problem.num_outputs
            self.model.addPoint(inputs, output)
        self.model.train()



    # The infill function will work automatically, since we have defined get_infill and update_model


    # we initialize and store a kriging model on the stored training data, and then
    # run the internal model's train function.
    def train(self):
        self.model = kriging(self.data.values[:,:-1], self.data.values[:,-1])
        self.model.train()

    # this model expects a 2d array representing a batch of inputs, but we only want
    # to evaluate one input point at a time, so we wrap the inputs in a list before passing
    # them to the model
    def eval_single(self, values):
        return ((self.model.predict(list(values)),), ())


@pytest.fixture
def building():
    return ef.get_building()


@pytest.fixture
def parameters():
    parameters=expand_plist(
    {'Mass NonRes Wall Insulation':
      {'Thickness': (0.05, 0.99)}
    })

    parameters.append(wwr())
    return parameters


def test_init(building, parameters):
    """Testing the initialization and functionality of an adaptive surrogate model"""

    problem = EPProblem(parameters)
    evaluator = EvaluatorEP(problem, building)

    inputs = sampling.dist_sampler(sampling.seeded_sampler, problem, 10)
    inputs = sampling.add_extremes(inputs, problem)

    #using the custom class for our own model training
    k = KirgingEval(reference=evaluator)
    k.do_infill(inputs)
    k.model.train()
    k.infill(5)

    #change this to 0 to see stdout and stderr
    assert 1



import pytest
import pandas as pd
from pandas import DataFrame as DF
import numpy as np

import eppy_funcs as ef
from evaluator import EvaluatorEP, EvaluatorSR
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector
from problem import EPProblem
import sampling

#Some code taken and adapted from example notebooks, made to work through pytest

@pytest.fixture
def building():
    return ef.get_building()


@pytest.fixture
def problem():
    problem = EPProblem([
        Parameter(
            FieldSelector(object_name='Mass NonRes Wall Insulation', field_name='Thickness'),
            RangeParameter(min_val = 0.01, max_val=0.99)),
        Parameter(
            FieldSelector(class_name='Construction', object_name='ext-slab', field_name='Outside Layer'),
            CategoryParameter(options=('HW CONCRETE', 'Invalid Material')))])
    return problem


@pytest.fixture
def samples(problem):
    #seeded_sampler results in indexes 0-3 being Invalid Material and 4 being HW CONCRETE
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, 5)
    return samples


def test_exception_throwing(building, problem, samples):
    """test to make sure exceptions are thrown or ignored depending on the error_mode"""

    #check that an exception is raised with FailFast mode
    with pytest.raises(Exception):
        evaluator = EvaluatorEP(problem, building, error_mode='FailFast').df_apply(samples)

    #check that no exceptions are raised in the other mode
    try:
        evaluator = EvaluatorEP(problem, building, error_mode='Silent').df_apply(samples)
    except Exception as e:
        assert 0

    #change this to 0 to see stdout and stderr
    assert 1


def test_error_values(building, problem, samples):
    """check that automatic error handling will assign the desired error values"""

    #check that the default error value is assigned to invalid materials
    evaluator = EvaluatorEP(problem, building, error_mode='Silent')
    results = evaluator.df_apply(samples)
    value = results.iloc[0]['Electricity:Facility']
    assert value == np.inf, f'Invalid material not assigned the default error value, value assigned was: {value}'

    #check that a custom error value is assigned to invalid materials
    error_value=((-1,), ())
    evaluator = EvaluatorEP(problem, building, error_mode='Silent', error_value=error_value)
    results = evaluator.df_apply(samples)
    value = results.iloc[0]['Electricity:Facility']
    assert value == -1, f'Invalid material not assigned the correct error value, value assigned was: {value}'

    #check that valid inputs aren't assigned error values
    value = results.iloc[4]['Electricity:Facility']
    assert value != -1, f'Valid material was assigned the error value: {error_value}'

    #change this to 0 to see stdout and stderr
    assert 1



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# Two types of outputs are supported: Objectives and Constraints. These are both made using the `MeterReader` and `VariableReader` classes. The only difference is how they are used by the problem.

# +
from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorEP
from besos.parameters import RangeParameter, wwr, FieldSelector, Parameter
from besos.objectives import MeterReader, VariableReader, clear_outputs
from besos.problem import EPProblem, Problem
from besos.optimizer import NSGAII

import pandas as pd
# -

building = ef.get_building()
clear_outputs(building)

# Objectives and constraints can be specified in one of 3 ways. The most direct is by calling their constructor.

objectives = [MeterReader(key_name='Electricity:Facility', class_name='Output:Meter', frequency='Hourly')]
EPProblem(outputs=objectives)

# The constructor has defaults, so we can often ommit `class_name` and `frequency`. Since the key name is often all that is needed, we can just use a list of the `key_names`. This list will be automatically be converted by `EPProblem`. Meters and variables that do not have a `frequency` specified will default to any frequency that is already used for that output, or if none is used yet then they will use Hourly.

objectives = ['Electricity:Facility']
EPProblem(outputs=objectives)

# If we do not need the output-reading features of meters, we can use `Problem` instead of `EPProblem`, and they will be converted to `Objective` objects which act as placeholders. `EPProblem` converts them to `Meter:Reader` objects. Either of these conversions can be overriden using the converters argument.  

objectives = ['any', 'names', 'work']
Problem(outputs=objectives)

# The `func` argument to objectives/constraints is used to aggrgate the individual measurements from the model. By default, all measurements are added together. If we wanted to instead minimize the variance, we would need to write our own aggrgation function

# +
def variance(result):
    return result.data['Value'].var()

objectives = [MeterReader('Electricity:Facility', name='Electricity Usage'),
              MeterReader('Electricity:Facility',func=variance, name='Electricity Variance')
             ]
# -

# When we want to specify the direction of optimisation, we can use `minmize_outputs`. The default is true for all objectives. Here we say we want to search for a design that has low but highly variable electricity use, and no more than 800 kg of CO2 emission. (The electricity use objectives were defined in the cell above)

# +
inputs = [wwr(), Parameter(FieldSelector(class_name='Material',
                                         object_name='Mass NonRes Wall Insulation',
                                         field_name='Thickness'),
                           RangeParameter(0.01, 0.99))
         ]

evaluator = EvaluatorEP(EPProblem(inputs=inputs,
                                  outputs=objectives, minimize_outputs=[True, False],
                                  constraints=['CO2:Facility'], constraint_bounds=['<=800']),
                        building, out_dir='../E_Plus_Files/')
# -

# this cell runs the optimisation, and may take a while.
NSGAII(evaluator, evaluations=100, population_size=20)


